// This module features so called "unsafe" code that can be used to improve performance.
// However it appears that it usage of it actually slowed down parser, so this "unsafe" code is not
// active by default. 
//#define UNSAFE_CODE

// if defined then HTMLparser will be compiled with fastest parsing code that in some very rare cases
// won't be 100% correct
#define FASTEST_PARSING

using System;
using System.IO;
using System.Collections;
using System.Text;

/*

Copyright (c) Alex Chudnovsky, Majestic-12 Ltd (UK). 2005+ All rights reserved
Web:		http://www.majestic12.co.uk
E-mail:		alexc@majestic12.co.uk

Redistribution and use in source and binary forms, with or without modification, are 
permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions 
		and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
		and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the Majestic-12 nor the names of its contributors may be used to endorse or 
		promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

/// <summary>
/// Description:	HTMLparser -- low level class that splits HTML into tokens 
///					such as tag, comments, text etc.
///					
///					Change source with great care - a lot of effort went into optimising it to make sure it performs
///					very well, for this reason some range checks were removed in cases when big enough buffers
///					were sufficient for 99.9999% of HTML pages out there.
///					
///					This does not mean you won't get an exception, so when you are parsing, then make sure you
///					catch exceptions from parser.
///					
/// Author:			Alex Chudnovsky <alexc@majestic12.co.uk>
/// 	
/// History:		
///							
///					04/12/05 v2.0.0 Second public release - lots of changes in between
///					 4/02/06 v1.0.6  Added fix to raw HTML not being stored, thanks
///									for this go to Christopher Woodill <chriswoodill@yahoo.com>
///					22/11/05 v1.0.5	
///									! Fixed issue with SCRIPT tag parsing for case when COMMENT tag was inside it
///					11/11/05 v1.0.4 
///									! Fixed problem with tag's param's value not using quotes at all with /' in it
///									  ($#%&-{^!s)
///									! Fixed problem with closed tag using / at the end of the tag rather than
///									  at the beginning (where its supposed to be?)
///									! Fixed problem with Unicode entities
///					02/11/05 v1.0.3 
///									+ Added custom scan until end of script tag to combat broken
///									  HTML with > used for comparison and confusing parser thinkings its a new tag
///									! Fixed bug with tag params starting with a quote
///									+ Decoding from non-default encodings is now possible (requires client support
///									  to set current encoding based on meta tags or headers)
///									! Fixed bug with > char present in tag's param right after quote char
///					28/10/05 v1.0.2 ! Fixed bug with whitespace in param values
///									+ Added small protection against bug in HTML due misplaced quotes
///					19/10/05 v1.0.1 
///									! Faster entity parsing
///									! Faster whitespace parsing
///									! For performance reasons HTML comments parsing moved into separate function
///									! Speculative loop unrolling in critical places
///									! Faster parsing mode (as a tradeoff for minor incorrectness)
///					01/10/05 v1.0.0	Public release
///					  ...			Many changes here
///					04/08/04 v0.5.0	New
///							 
/// 
/// 
/// </summary>

#if CONAN_NAMESPACE
using MJ12ConAn;

namespace MJ12ConAn
#else
using Majestic12;

namespace Majestic12
#endif
{

	/// <summary>
	/// Type of parsed HTML chunk (token)
	/// </summary>
	public enum HTMLchunkType
	{
		/// <summary>
		/// Text data from HTML
		/// </summary>
		Text		= 0,

		/// <summary>
		/// Open tag
		/// </summary>
		OpenTag		= 1,

		/// <summary>
		/// Closed tag
		/// </summary>
		CloseTag	= 2,

		/// <summary>
		/// Data between HTML comments: requires manual call to HTMLchunk.Finalise() before data is available
		/// </summary>
		Comment		= 3,

		/// <summary>
		/// Data betweeen script tags:  requires manual call to HTMLchunk.Finalise() before data is available
		/// </summary>
		Script		= 4,
	};

	/// <summary>
	/// Class for fast dynamic string building - it is faster than StringBuilder
	/// </summary>
#if UNSAFE_CODE
	///<exclude/>
	public unsafe class DynaString
#else
	///<exclude/>
	public class DynaString	: IDisposable
#endif
	{
		/// <summary>
		/// Finalised text will be available in this string
		/// </summary>
		public string sText;

		/// <summary>
		/// CRITICAL: that much capacity will be allocated (once) for this object -- for performance reasons
		/// we do NOT have range checks because we make reasonably safe assumption that accumulated string will
		/// fit into the buffer. If you have very abnormal strings then you should increase buffer accordingly.
		/// </summary>
		public static int TEXT_CAPACITY=1024*256-1;

		public byte[] bBuffer;
		public int iBufPos;
		private int Length;

		Encoding oEnc=Encoding.Default;

		private bool bDisposed=false;

		/// <summary>
		/// Constructor 
		/// </summary>
		/// <param name="sString">Initial string</param>
		public DynaString(string sString)
		{
			sText=sString;
			iBufPos=0;
			bBuffer=new byte[TEXT_CAPACITY+1];
			Length=sString.Length;
		}

		/// <summary>
		/// Converts data in buffer to string
		/// </summary>
		/// <returns>String</returns>
		public override string ToString()
		{
			if(iBufPos>0)
				Finalise();
			
			return sText;
		}

#if USE_DESTRUCTORS
		~DynaString()
		{
			Dispose(false);
		}
#endif
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool bDisposing)
		{
			if(!bDisposed)
			{
			   bBuffer=null;
			}

			bDisposed=true;
		}

		/// <summary>
		/// Lower cases data in buffer and returns as string. 
		/// WARNING: This is only to be used for ASCII lowercasing - HTML params and tags, not their values
		/// </summary>
		/// <returns>String that is now accessible via sText</returns>
		public unsafe string ToLowerString()
		{
			if(sText.Length==0)
			{	

#if UNSAFE_CODE || true
				// lower case it directly in the buffer
				fixed(byte* pBuffer=bBuffer)
				{
					int i=0; 
					for(byte *p=pBuffer; i<iBufPos; i++, p++)
					{
						if(*p>=65 && *p<=90)
							*p=(byte)(*p+32);
					}
					
				}

				ToString();
#else
			ToString();
			sText=sText.ToLower();
#endif
			}
			else
			{
				ToString();
			
#if UNSAFE_CODE			
			// fast in place conversion to lower case text

			fixed (char *pfixed = sText)
				for (char *p=pfixed; *p != 0; p++)
					*p = char.ToLower(*p);
#else
			sText=sText.ToLower();
#endif
			}

			return sText;
		}

		/// <summary>
		/// Resets object to zero length string
		/// </summary>
		public void Clear()
		{
			sText="";
			Length=0;
			iBufPos=0;
		}


		/// <summary>
		/// Sets encoding to be used for conversion of binary data into string
		/// </summary>
		/// <param name="p_oEnc">Encoding object</param>
		public void SetEncoding(Encoding p_oEnc)
		{
			oEnc=p_oEnc;
		}

		/// <summary>
		/// Appends a "char" to the buffer
		/// </summary>
		/// <param name="cChar">Appends char (byte really)</param>
		public unsafe void Append(byte cChar)
		{
			// Length++;

			if(iBufPos>=TEXT_CAPACITY)
			{
				if(sText.Length==0)
				{
#if UNSAFE_CODE || false
					fixed(byte* pBuffer=bBuffer)
					{
						sText=new String((sbyte*)pBuffer,0,iBufPos,System.Text.Encoding.Default);
					}
#else	
					
					sText=oEnc.GetString(bBuffer,0,iBufPos);
#endif
				}
				else
					//sText+=new string(bBuffer,0,iBufPos);
					sText+=oEnc.GetString(bBuffer,0,iBufPos);

				Length+=iBufPos;

				iBufPos=0;
			}

			bBuffer[iBufPos++]=cChar;
		}
		
		/// <summary>
		/// Internal finaliser - creates string from data accumulated in buffer
		/// </summary>
		private void Finalise()
		{
			if(iBufPos>0)
			{
				if(sText.Length==0)
				{
#if UNSAFE_CODE	
					fixed(byte* pBuffer=bBuffer)
					{
						sText=new String((sbyte*)pBuffer,0,iBufPos,System.Text.Encoding.Default);
					}
#else
					sText=oEnc.GetString(bBuffer,0,iBufPos);
#endif
				}
				else
					//sText+=new string(bBuffer,0,iBufPos);
					sText+=oEnc.GetString(bBuffer,0,iBufPos);

				Length+=iBufPos;
				iBufPos=0;
			}
		}
		

	}

	/// <summary>
	/// Parsed HTML token that is either text, comment, script, open or closed tag as indicated by the oType variable.
	/// </summary>
#if UNSAFE_CODE || true
	public unsafe class HTMLchunk : IDisposable
#else
	public class HTMLchunk
#endif
	{	
		/// <summary>
		/// Maximum default capacity of buffer that will keep data
		/// </summary>
		/// <exclude/>
		public static int TEXT_CAPACITY=1024*256;

		/// <summary>
		/// Maximum number of parameters in a tag - should be high enough to fit most
		/// </summary>
		/// <exclude/>
		public static int MAX_PARAMS=256;	

		/// <summary>
		/// Chunk type showing whether its text, open or close tag, comments or script.
		/// WARNING: if type is comments or script then you have to manually call Finalise(); method
		/// in order to have actual text of comments/scripts in oHTML variable
		/// </summary>
		public HTMLchunkType oType;

		/// <summary>
		/// 
		/// </summary>
		/// <exclude/>
		public byte[] bBuffer=new byte[TEXT_CAPACITY+1];

		/// <summary>
		/// 
		/// </summary>
		/// <exclude/>
		public int iBufPos=0;

		/// <summary>
		/// 
		/// </summary>
		/// <exclude/>
		public int iHTMLen=0;

		/// <summary>
		/// If true then tag params will be kept in a hash rather than in a fixed size arrays. 
		/// This will be slow down parsing, but make it easier to use.
		/// </summary>
		public bool bHashMode=true;

		/// <summary>
		/// For TAGS: it stores raw HTML that was parsed to generate thus chunk will be here UNLESS
		/// HTMLparser was configured not to store it there as it can improve performance
		/// <p>
		/// For TEXT or COMMENTS: actual text or comments - you MUST call Finalise(); first.
		/// </p>
		/// </summary>
		public string oHTML="";

		/// <summary>
		/// If its open/close tag type then this is where lowercased Tag will be kept
		/// </summary>
		public string sTag="";

		/// <summary>
		/// 
		/// </summary>
		/// <exclude/>
		public bool bClosure=false;

		/// <summary>
		/// 
		/// </summary>
		/// <exclude/>
		public bool bComments=false;
		
		/// <summary>
		/// True if entities were present (and transformed) in the original HTML
		/// </summary>
		/// <exclude/>
		public bool bEntities=false;

		/// <summary>
		/// Set to true if &lt; entity (tag start) was found 
		/// </summary>
		/// <exclude/>
 		public bool bLtEntity=false;

		/// <summary>
		/// Hashtable with tag parameters: keys are param names and values are param values.
		/// ONLY used if bHashMode is set to TRUE.
		/// </summary>
		public Hashtable oParams=null; 

		/// <summary>
		/// Number of parameters and values stored in sParams array.
		/// ONLY used if bHashMode is set to FALSE.
		/// </summary>
		public int iParams=0;

		/// <summary>
		/// Param names will be stored here - actual number is in iParams.
		/// ONLY used if bHashMode is set to FALSE.
		/// </summary>
		public string[] sParams=new string[MAX_PARAMS];

		/// <summary>
		/// Param values will be stored here - actual number is in iParams.
		/// ONLY used if bHashMode is set to FALSE.
		/// </summary>
		public string[] sValues=new string[MAX_PARAMS];

		/// <summary>
		/// Encoding to be used for conversion of binary data into strings
		/// </summary>
		Encoding oEnc=Encoding.Default;

		bool bDisposed=false;

		/// <summary>
		/// This function will convert parameters stored in sParams/sValues arrays into oParams hash
		/// Useful if generally parsing is done when bHashMode is FALSE. Hash operations are not the fastest, so
		/// its best not to use this function.
		/// </summary>
		public void ConvertParamsToHash()
		{
			if(oParams!=null)
				oParams.Clear();
			else
				oParams=new Hashtable();

			for(int i=0; i<iParams; i++)
			{
				oParams[sParams[i]]=sValues[i];
			}

		}

		/// <summary>
		/// Sets encoding to be used for conversion of binary data into string
		/// </summary>
		/// <param name="p_oEnc">Encoding object</param>
		public void SetEncoding(Encoding p_oEnc)
		{
			oEnc=p_oEnc;
		}

		/// <summary>
		/// Generates HTML based on current chunk's data 
		/// Note: this is not a high performance method
		/// </summary>
		/// <returns>HTML equivalent of this chunk</returns>
		public string GenerateHTML()
		{
			string sHTML="";

			switch(oType)
			{
					// matched open tag, ie <a href="">
				case HTMLchunkType.OpenTag:
					sHTML+="<"+sTag;

					if(iParams>0)
						sHTML+=" "+GenerateParamsHTML();

					sHTML+=">";

					break;

					// matched close tag, ie </a>
				case HTMLchunkType.CloseTag:

					sHTML+="</"+sTag;

					if(iParams>0)
						sHTML+=" "+GenerateParamsHTML();

					sHTML+=">";

					break;

				case HTMLchunkType.Script:
					Finalise();
					sHTML="<script>"+oHTML+"</script>";
					break;

				case HTMLchunkType.Comment:
					Finalise();
					sHTML="<!--"+oHTML+" -->";
					break;

					// matched normal text
				case HTMLchunkType.Text:

					return oHTML;
			};

			return sHTML;
		}

		/// <summary>
		/// Returns value of a parameter
		/// </summary>
		/// <param name="sParam">Parameter</param>
		/// <returns>Parameter value or empty string</returns>
		public string GetParamValue(string sParam)
		{
			if(!bHashMode)
			{			
				for(int i=0; i<iParams; i++)
				{
					if(sParams[i]==sParam)
						return sValues[i];
				}
			}
			else
			{
				object oValue=oParams[sParam];

				if(oValue!=null)
					return (string)oValue;
			}

			return "";
		}

		/// <summary>
		/// Generates HTML for params in this chunk
		/// </summary>
		/// <returns>String with HTML corresponding to params</returns>
		public string GenerateParamsHTML()
		{
			string sParamHTML="";

			if(bHashMode)
			{
				if(oParams.Count>0)
				{
					foreach(string sParam in oParams.Keys)
					{
						string sValue=oParams[sParam].ToString();

						if(sParamHTML.Length>0)
							sParamHTML+=" ";

						sParamHTML+=GenerateParamHTML(sParam,sValue);
					}

				}
			}
			else
			{
				// this is alternative method of getting params -- it may look less convinient
				// but it saves a LOT of CPU ticks while parsing. It makes sense when you only need
				// params for a few
				if(iParams>0)
				{
					for(int i=0; i<iParams; i++)
					{
						if(sParamHTML.Length>0)
							sParamHTML+=" ";

						sParamHTML+=GenerateParamHTML(sParams[i],sValues[i]);
					}

				}
						
			}

			return sParamHTML;
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool bDisposing)
		{
			if(!bDisposed)
			{
				if(oParams!=null)
					oParams=null;
			
				sParams=null;
				sValues=null;

				bBuffer=null;
			}

			bDisposed=true;
		}

		/// <summary>
		/// Generates HTML for param/value pair
		/// </summary>
		/// <param name="sParam">Param</param>
		/// <param name="sValue">Value (empty if not specified)</param>
		/// <returns>String with HTML</returns>
		public string GenerateParamHTML(string sParam,string sValue)
		{
			if(sValue.Length>0)
			{
				// check param's value for whitespace or quote chars, if they are not present, then
				// we can save 2 bytes by not generating quotes
				if(sValue.Length>20)
					return sParam+"='"+MakeSafeParamValue(sValue,'\'')+"'";

				for(int i=0; i<sValue.Length; i++)
				{
					switch(sValue[i])
					{
						case ' ':
						case '\t':
						case '\'':
						case '\"':
						case '\n':
						case '\r':
							return sParam+"='"+MakeSafeParamValue(sValue,'\'')+"'";

						default:
							break;
					};
				}

				return sParam+"="+sValue;
			}
			else
				return sParam;
		}

		/// <summary>
		/// Makes parameter value safe to be used in param - this will check for any conflicting quote chars,
		/// but not full entity-encoding
		/// </summary>
		/// <param name="sLine">Line of text</param>
		/// <param name="cQuoteChar">Quote char used in param - any such chars in text will be entity-encoded</param>
		/// <returns>Safe text to be used as param's value</returns>
		public static string MakeSafeParamValue(string sLine,char cQuoteChar)
		{
			// we speculatievly expect that in most cases we don't actually need to entity-encode string,

			for(int i=0; i<sLine.Length; i++)
			{
				if(sLine[i]==cQuoteChar)
				{
					// have to restart here
					StringBuilder oSB=new StringBuilder(sLine.Length+10);

					oSB.Append(sLine.Substring(0,i));

					for(int j=i; j<sLine.Length; j++)
					{
						char cChar=sLine[j];

						if(cChar==cQuoteChar)
							oSB.Append("&#"+((int)cChar).ToString()+";");
						else
							oSB.Append(cChar);
					}

					return oSB.ToString();
				}
			}
			

			return sLine;
		}

		/// <summary>
		/// Adds tag parameter to the chunk
		/// </summary>
		/// <param name="sParam">Parameter name (ie color)</param>
		/// <param name="sValue">Value of the parameter (ie white)</param>
		public void AddParam(string sParam,string sValue)
		{
			if(!bHashMode)
			{

				if(iParams<MAX_PARAMS)
				{
					sParams[iParams]=sParam;
					sValues[iParams]=sValue;

					iParams++;
				}
			}
			else
			{
				oParams[sParam]=sValue;
			}

		}

#if USE_DESTRUCTORS
		~HTMLchunk()
		{
			Dispose();
		}
#endif

		/// <summary>
		/// Clears chunk preparing it for 
		/// </summary>
		public void Clear()
		{
			iHTMLen=iBufPos=0;
			//oHTML=null;
			sTag=oHTML="";
			//sTag=null;
			//sTag="";
			bLtEntity=bEntities=bComments=bClosure=false;
			
			/*
			bComments=false;
			bEntities=false;
			bLtEntity=false;
			*/

			if(!bHashMode)
			{
				/*
				for(int i=0; i<iParams; i++)
				{
					sParams[i]=null;
					sValues[i]=null;
				}
				*/

				iParams=0;
			}
			else
			{
				if(oParams!=null)
					oParams.Clear();
			}
			//if(oParams.Count>0)
			//	oParams=new Hashtable();
		}

		/// <summary>
		/// Initialises new HTMLchunk
		/// </summary>
		/// <param name="p_bHashMode">Sets <seealso cref="bHashMode"/></param>
		public HTMLchunk(bool p_bHashMode)
		{
			bHashMode=p_bHashMode;
			
			if(bHashMode)
				oParams=new Hashtable();
		}

		/// <summary>
		/// Appends char to chunk
		/// </summary>
		/// <param name="cChar">Char (byte really)</param>
		public void Append(byte cChar)
		{
			// no range check here for performance reasons - this function gets called VERY often

			bBuffer[iBufPos++]=cChar;
		}

		/// <summary>
		/// Appends proper char with smart handling of Unicode chars
		/// </summary>
		/// <param name="cChar">Char to append</param>
		public void Append(char cChar)
		{
			if(cChar<=127)
				bBuffer[iBufPos++]=(byte)cChar;
			else
			{
				// unicode character
				byte[] bBytes=oEnc.GetBytes(cChar.ToString());

				for(int i=0; i<bBytes.Length; i++)
					bBuffer[iBufPos++]=bBytes[i];
			}
		}

		/// <summary>
		/// Finalises data from buffer into HTML string
		/// </summary>
		public void Finalise()
		{
			if(oHTML.Length==0)
			{
				if(iBufPos>0)
				{
#if UNSAFE_CODE
					fixed(byte* pBuffer=bBuffer)
					{
						oHTML=new String((sbyte*)pBuffer,0,iBufPos,System.Text.Encoding.Default);
					}
#else
					//oHTML=Encoding.Default.GetString(bBuffer,0,iBufPos);
					oHTML=oEnc.GetString(bBuffer,0,iBufPos);
#endif
				}
			}
			else
			{
				if(iBufPos>0)
					//oHTML+=new string(bBuffer,0,iBufPos);
					//oHTML+=Encoding.Default.GetString(bBuffer,0,iBufPos);
					oHTML+=oEnc.GetString(bBuffer,0,iBufPos);
			}

			iHTMLen+=iBufPos;
			iBufPos=0;
		}

	}

	/// <summary>
	/// Allows to parse HTML by splitting it into small token (HTMLchunks) such as tags, text, comments etc.
	/// </summary>
	public class HTMLparser : IDisposable
	{
		/// <summary>
		/// If false then HTML entities (like "nbsp") will not be decoded
		/// </summary>
		public bool bDecodeEntities=false;
		
		/// <summary>
		/// If true then parsed tag chunks will contain raw HTML, otherwise only comments will have it
		/// <p>
		/// Performance hint: keep it as false
		/// </p>
		/// </summary>
		/// <exclude/>
		public bool bKeepRawHTML=false;

		/// <summary>
		/// If true then pure whitespace before tags will be ignored - but only IF its purely whitespace. 
		/// Enabling this feature will increase performance, however this will be at cost of correctness as 
		/// some text has essential whitespacing done just before tags.
		/// </summary>
		public bool bIgnoreWhiteSpaceBeforeTags=false;

		/// <summary>
		/// Internal -- dynamic string for text accumulation
		/// </summary>
		DynaString sText=new DynaString("");

		/// <summary>
		/// This chunk will be returned when it was parsed
		/// </summary>
		HTMLchunk oChunk=new HTMLchunk(true);

		// if true nested comments (<!-- <!-- --> -->) will be understood
		//bool bNestedComments=false;

		//byte[] cHTML=null;

		/// <summary>
		/// Encoding used to convert binary data into string
		/// </summary>
		Encoding oEnc=null;

		/// <summary>
		/// Byte array with HTML will be kept here
		/// </summary>
		byte[] bHTML;

		/// <summary>
		/// Current position pointing to byte in bHTML
		/// </summary>
		/// <exclude/
		public int iCurPos=0;

		/// <summary>
		/// Length of bHTML -- it appears to be faster to use it than bHTML.Length
		/// </summary>
		int iDataLength=0;

		/// <summary>
		/// Supported HTML entities
		/// </summary>
		FastHash oEntities=null;

		/// <summary>
		/// Supported HTML entities
		/// </summary>
		static FastHash oAllEntities=null;

		/// <summary>
		/// If false then text will be ignored -- it will make parser run faster but in effect only HTML tags 
		/// will be returned, its in effect only useful when you just want to parse links or something like this.
		/// </summary>
		public bool bTextMode=true;

		/// <summary>
		/// Internal heuristics for entiries: these will be set to min and max string lengths of known HTML entities
		/// </summary>
		int iMinEntityLen=0,iMaxEntityLen=0;

		static int iAllMinEntityLen=0,iAllMaxEntityLen=0;

		/// <summary>
		/// Array to provide reverse lookup for entities
		/// </summary>
		static string[] sEntityReverseLookup;
			

		/// <summary>
		/// Whitespace lookup table - 0 is not whitespace, otherwise it is
		/// </summary>
		byte[] bWhiteSpace=new byte[byte.MaxValue+1];

		/// <summary>
		/// If true then only minimal set of entities will be parsed, everything else including numbers based
		/// entities will be returned as is. This is useful for when HTML content needs to be extracted with subsequent parsing, in this case resolution of entities will be a problem
		/// </summary>
		bool bMiniEntities=false;

		static HTMLparser()
		{
			oAllEntities=InitEntities(ref iAllMinEntityLen,ref iAllMaxEntityLen,out sEntityReverseLookup);
		}

		public HTMLparser()
		{
			oEntities=InitEntities(ref iMinEntityLen,ref iMaxEntityLen,out sEntityReverseLookup);	
			bMiniEntities=false;


			bWhiteSpace[' ']=1;
			bWhiteSpace['\t']=1;
			bWhiteSpace[13]=1;
			bWhiteSpace[10]=1;
		}

		/// <summary>
		/// Sets decode entities mode
		/// </summary>
		/// <param name="bMode">If true then entities will be converted into chars, otherwise left as is</param>
		public void SetDecodeEntitiesMode(bool bMode)
		{
			bDecodeEntities=bMode;
		}

		/// <summary>
		/// Sets chunk param hash mode
		/// </summary>
		/// <param name="bHashMode">If true then tag's params will be kept in Chunk's hashtable (slower), otherwise kept in arrays (sParams/sValues)</param>
		public void SetChunkHashMode(bool bHashMode)
		{
			oChunk.bHashMode=bHashMode;
		}

#if USE_DESTRUCTORS
		~HTMLparser()
		{
			Dispose(false);	
		}
#endif

		bool bDisposed=false;

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool bDisposing)
		{
			if(!bDisposed)
			{
				oChunk.Dispose();
				sText.Dispose();

				bHTML=null;
				oEntities=null;
			}

			bDisposed=true;
		}


		/// <summary>
		/// Closes object and releases all allocated resources
		/// </summary>
		public void Close()
		{
			if(bHTML!=null)
				bHTML=null;

			if(oEntities!=null)
			{
				oEntities.Clear();
				oEntities=null;
			}
		}

		/// <summary>
		/// Sets encoding 
		/// </summary>
		/// <param name="p_oEnc">Encoding object</param>
		public void SetEncoding(Encoding p_oEnc)
		{
			oChunk.SetEncoding(p_oEnc);
			sText.SetEncoding(p_oEnc);
			
			if(p_oEnc!=null)
				oEnc=p_oEnc;
			/*
			if(oCharset.IsSupported(sCharset))
			{
				oCharset.SetCharset(sCharset);
				//bCharConv=true;
			}
			//else Console.WriteLine("Charset '{0}' is not supported!",sCharset);
			*/
		}

		/// <summary>
		/// Sets current encoding in format used in HTTP headers and HTML META tags
		/// </summary>
		/// <param name="sCharSet">Charset as </param>
		/// <returns>True if encoding was set, false otherwise (in which case Default encoding will be used)</returns>
		public bool SetEncoding(string sCharSet)
		{
			try
			{
				if(sCharSet.IndexOf(";")!=-1)
					sCharSet=GetCharSet(sCharSet);

				sCharSet=sCharSet.Replace("?","");

				oEnc=Encoding.GetEncoding(sCharSet);

				// FIXIT: check here that the encoding set was not some fallback to default
				// due to lack of support on target machine

				//Utils.WriteLine("Setting encoding: "+sCharSet);
				SetEncoding(oEnc);
			}
			catch(Exception oEx)
			{
				// encoding was not supported - we will fall back to default one
				oEnc=Encoding.Default;
				SetEncoding(oEnc);
		
				return false;
			}

			return true;
		}

		/// <summary>
		/// Retrieves charset information from format used in HTTP headers and META descriptions
		/// </summary>
		/// <param name="sData">Data to find charset info from</param>
		/// <returns>Charset</returns>
		static string GetCharSet(string sData)
		{
			try
			{
				if(sData==null)
					return "";

				int iIdx=sData.ToLower().IndexOf("charset=");

				if(iIdx!=-1)
					return sData.Substring(iIdx+8,sData.Length-iIdx-8).ToLower().Trim();
				
			}
			catch(Exception oEx)
			{
				Console.WriteLine(oEx.ToString());
			}

			return "";
		}

		/// <summary>
		/// Inits mini-entities mode: only "nbsp" will be converted into space, all other entities 
		/// will be left as is
		/// </summary>
		public void InitMiniEntities()
		{
			oEntities=new FastHash();

			oEntities.Add("nbsp",32);
			bMiniEntities=true;
		}

		/// <summary>
		/// Initialises list of entities
		/// </summary>
		private static FastHash InitEntities(ref int iMinEntityLen,ref int iMaxEntityLen,out string[] sEntityReverseLookup)
		{
			FastHash oEntities=new FastHash();

			// FIXIT: we will treat non-breakable space... as space!?!
			// perhaps it would be better to have separate return types for entities?
			oEntities.Add("nbsp",32); //oEntities.Add("nbsp",160);
			oEntities.Add("iexcl",161);
			oEntities.Add("cent",162);
			oEntities.Add("pound",163);
			oEntities.Add("curren",164);
			oEntities.Add("yen",165);
			oEntities.Add("brvbar",166);
			oEntities.Add("sect",167);
			oEntities.Add("uml",168);
			oEntities.Add("copy",169);
			oEntities.Add("ordf",170);
			oEntities.Add("laquo",171);
			oEntities.Add("not",172);
			oEntities.Add("shy",173);
			oEntities.Add("reg",174);
			oEntities.Add("macr",175);
			oEntities.Add("deg",176);
			oEntities.Add("plusmn",177);
			oEntities.Add("sup2",178);
			oEntities.Add("sup3",179);
			oEntities.Add("acute",180);
			oEntities.Add("micro",181);
			oEntities.Add("para",182);
			oEntities.Add("middot",183);
			oEntities.Add("cedil",184);
			oEntities.Add("sup1",185);
			oEntities.Add("ordm",186);
			oEntities.Add("raquo",187);
			oEntities.Add("frac14",188);
			oEntities.Add("frac12",189);
			oEntities.Add("frac34",190);
			oEntities.Add("iquest",191);
			oEntities.Add("Agrave",192);
			oEntities.Add("Aacute",193);
			oEntities.Add("Acirc",194);
			oEntities.Add("Atilde",195);
			oEntities.Add("Auml",196);
			oEntities.Add("Aring",197);
			oEntities.Add("AElig",198);
			oEntities.Add("Ccedil",199);
			oEntities.Add("Egrave",200);
			oEntities.Add("Eacute",201);
			oEntities.Add("Ecirc",202);
			oEntities.Add("Euml",203);
			oEntities.Add("Igrave",204);
			oEntities.Add("Iacute",205);
			oEntities.Add("Icirc",206);
			oEntities.Add("Iuml",207);
			oEntities.Add("ETH",208);
			oEntities.Add("Ntilde",209);
			oEntities.Add("Ograve",210);
			oEntities.Add("Oacute",211);
			oEntities.Add("Ocirc",212);
			oEntities.Add("Otilde",213);
			oEntities.Add("Ouml",214);
			oEntities.Add("times",215);
			oEntities.Add("Oslash",216);
			oEntities.Add("Ugrave",217);
			oEntities.Add("Uacute",218);
			oEntities.Add("Ucirc",219);
			oEntities.Add("Uuml",220);
			oEntities.Add("Yacute",221);
			oEntities.Add("THORN",222);
			oEntities.Add("szlig",223);
			oEntities.Add("agrave",224);
			oEntities.Add("aacute",225);
			oEntities.Add("acirc",226);
			oEntities.Add("atilde",227);
			oEntities.Add("auml",228);
			oEntities.Add("aring",229);
			oEntities.Add("aelig",230);
			oEntities.Add("ccedil",231);
			oEntities.Add("egrave",232);
			oEntities.Add("eacute",233);
			oEntities.Add("ecirc",234);
			oEntities.Add("euml",235);
			oEntities.Add("igrave",236);
			oEntities.Add("iacute",237);
			oEntities.Add("icirc",238);
			oEntities.Add("iuml",239);
			oEntities.Add("eth",240);
			oEntities.Add("ntilde",241);
			oEntities.Add("ograve",242);
			oEntities.Add("oacute",243);
			oEntities.Add("ocirc",244);
			oEntities.Add("otilde",245);
			oEntities.Add("ouml",246);
			oEntities.Add("divide",247);
			oEntities.Add("oslash",248);
			oEntities.Add("ugrave",249);
			oEntities.Add("uacute",250);
			oEntities.Add("ucirc",251);
			oEntities.Add("uuml",252);
			oEntities.Add("yacute",253);
			oEntities.Add("thorn",254);
			oEntities.Add("yuml",255);
			oEntities.Add("quot",34);
			oEntities.Add("amp",38);
			oEntities.Add("lt",60);
			oEntities.Add("gt",62);

			// unicode supported by default
			if(true)
			{
				oEntities.Add("OElig",338);
				oEntities.Add("oelig",339);
				oEntities.Add("Scaron",352);
				oEntities.Add("scaron",353);
				oEntities.Add("Yuml",376);
				oEntities.Add("circ",710);
				oEntities.Add("tilde",732);
				oEntities.Add("ensp",8194);
				oEntities.Add("emsp",8195);
				oEntities.Add("thinsp",8201);
				oEntities.Add("zwnj",8204);
				oEntities.Add("zwj",8205);
				oEntities.Add("lrm",8206);
				oEntities.Add("rlm",8207);
				oEntities.Add("ndash",8211);
				oEntities.Add("mdash",8212);
				oEntities.Add("lsquo",8216);
				oEntities.Add("rsquo",8217);
				oEntities.Add("sbquo",8218);
				oEntities.Add("ldquo",8220);
				oEntities.Add("rdquo",8221);
				oEntities.Add("bdquo",8222);
				oEntities.Add("dagger",8224);
				oEntities.Add("Dagger",8225);
				oEntities.Add("permil",8240);
				oEntities.Add("lsaquo",8249);
				oEntities.Add("rsaquo",8250);
				oEntities.Add("euro",8364);
				oEntities.Add("fnof",402);
				oEntities.Add("Alpha",913);
				oEntities.Add("Beta",914);
				oEntities.Add("Gamma",915);
				oEntities.Add("Delta",916);
				oEntities.Add("Epsilon",917);
				oEntities.Add("Zeta",918);
				oEntities.Add("Eta",919);
				oEntities.Add("Theta",920);
				oEntities.Add("Iota",921);
				oEntities.Add("Kappa",922);
				oEntities.Add("Lambda",923);
				oEntities.Add("Mu",924);
				oEntities.Add("Nu",925);
				oEntities.Add("Xi",926);
				oEntities.Add("Omicron",927);
				oEntities.Add("Pi",928);
				oEntities.Add("Rho",929);
				oEntities.Add("Sigma",931);
				oEntities.Add("Tau",932);
				oEntities.Add("Upsilon",933);
				oEntities.Add("Phi",934);
				oEntities.Add("Chi",935);
				oEntities.Add("Psi",936);
				oEntities.Add("Omega",937);
				oEntities.Add("alpha",945);
				oEntities.Add("beta",946);
				oEntities.Add("gamma",947);
				oEntities.Add("delta",948);
				oEntities.Add("epsilon",949);
				oEntities.Add("zeta",950);
				oEntities.Add("eta",951);
				oEntities.Add("theta",952);
				oEntities.Add("iota",953);
				oEntities.Add("kappa",954);
				oEntities.Add("lambda",955);
				oEntities.Add("mu",956);
				oEntities.Add("nu",957);
				oEntities.Add("xi",958);
				oEntities.Add("omicron",959);
				oEntities.Add("pi",960);
				oEntities.Add("rho",961);
				oEntities.Add("sigmaf",962);
				oEntities.Add("sigma",963);
				oEntities.Add("tau",964);
				oEntities.Add("upsilon",965);
				oEntities.Add("phi",966);
				oEntities.Add("chi",967);
				oEntities.Add("psi",968);
				oEntities.Add("omega",969);
				oEntities.Add("thetasym",977);
				oEntities.Add("upsih",978);
				oEntities.Add("piv",982);
				oEntities.Add("bull",8226);
				oEntities.Add("hellip",8230);
				oEntities.Add("prime",8242);
				oEntities.Add("Prime",8243);
				oEntities.Add("oline",8254);
				oEntities.Add("frasl",8260);
				oEntities.Add("weierp",8472);
				oEntities.Add("image",8465);
				oEntities.Add("real",8476);
				oEntities.Add("trade",8482);
				oEntities.Add("alefsym",8501);
				oEntities.Add("larr",8592);
				oEntities.Add("uarr",8593);
				oEntities.Add("rarr",8594);
				oEntities.Add("darr",8595);
				oEntities.Add("harr",8596);
				oEntities.Add("crarr",8629);
				oEntities.Add("lArr",8656);
				oEntities.Add("uArr",8657);
				oEntities.Add("rArr",8658);
				oEntities.Add("dArr",8659);
				oEntities.Add("hArr",8660);
				oEntities.Add("forall",8704);
				oEntities.Add("part",8706);
				oEntities.Add("exist",8707);
				oEntities.Add("empty",8709);
				oEntities.Add("nabla",8711);
				oEntities.Add("isin",8712);
				oEntities.Add("notin",8713);
				oEntities.Add("ni",8715);
				oEntities.Add("prod",8719);
				oEntities.Add("sum",8721);
				oEntities.Add("minus",8722);
				oEntities.Add("lowast",8727);
				oEntities.Add("radic",8730);
				oEntities.Add("prop",8733);
				oEntities.Add("infin",8734);
				oEntities.Add("ang",8736);
				oEntities.Add("and",8743);
				oEntities.Add("or",8744);
				oEntities.Add("cap",8745);
				oEntities.Add("cup",8746);
				oEntities.Add("int",8747);
				oEntities.Add("there4",8756);
				oEntities.Add("sim",8764);
				oEntities.Add("cong",8773);
				oEntities.Add("asymp",8776);
				oEntities.Add("ne",8800);
				oEntities.Add("equiv",8801);
				oEntities.Add("le",8804);
				oEntities.Add("ge",8805);
				oEntities.Add("sub",8834);
				oEntities.Add("sup",8835);
				oEntities.Add("nsub",8836);
				oEntities.Add("sube",8838);
				oEntities.Add("supe",8839);
				oEntities.Add("oplus",8853);
				oEntities.Add("otimes",8855);
				oEntities.Add("perp",8869);
				oEntities.Add("sdot",8901);
				oEntities.Add("lceil",8968);
				oEntities.Add("rceil",8969);
				oEntities.Add("lfloor",8970);
				oEntities.Add("rfloor",8971);
				oEntities.Add("lang",9001);
				oEntities.Add("rang",9002);
				oEntities.Add("loz",9674);
				oEntities.Add("spades",9824);
				oEntities.Add("clubs",9827);
				oEntities.Add("hearts",9829);
				oEntities.Add("diams",9830);
			}

			sEntityReverseLookup=new string[10000];
				
			// calculate min/max lenght of known entities
			foreach(string sKey in oEntities.Keys)
			{
				if(sKey.Length<iMinEntityLen || iMinEntityLen==0)
					iMinEntityLen=sKey.Length;

				if(sKey.Length>iMaxEntityLen || iMaxEntityLen==0)
					iMaxEntityLen=sKey.Length;

				// remember key at given offset
				sEntityReverseLookup[(int)oEntities[sKey]]=sKey;
			}

			// we don't want to change spaces
			sEntityReverseLookup[32]=null;

			return oEntities;
		}
		
		/// <summary>
		/// Parses line and changes known entiry characters into proper HTML entiries
		/// </summary>
		/// <param name="sLine">Line of text</param>
		/// <returns>Line of text with proper HTML entities</returns>
		public string ChangeToEntities(string sLine)
		{
			try
			{
				// scan string first and if 
				for(int i=0; i<sLine.Length; i++)
				{				
					int cChar=(int)sLine[i];

					// yeah I know its lame but its 3:30am and I had v.long debugging session :-/
					switch(cChar)
					{
						case 39:
						case 145:
						case 146:
						case 147:
						case 148:
							return intChangeToEntities(sLine,i);

						default:
							break;
					};

					if(cChar<sEntityReverseLookup.Length && sEntityReverseLookup[cChar]!=null)
						return intChangeToEntities(sLine,i);
				}

			}
			catch(Exception oEx)
			{
				Console.WriteLine("Entity exception: "+oEx.ToString());
			}

			// nothing need to be changed
			return sLine;
		}

		/// <summary>
		/// Parses line and changes known entiry characters into proper HTML entiries
		/// </summary>
		/// <param name="sLine">Line of text</param>
		/// <param name="iFrom">Char from which scanning should start</param>
		/// <returns>Line of text with proper HTML entities</returns>
		private string intChangeToEntities(string sLine,int iFrom)
		{
			StringBuilder oSB=new StringBuilder(sLine.Length);

			if(iFrom>0)
				oSB.Append(sLine.Substring(0,iFrom));

			for(int i=iFrom; i<sLine.Length; i++)
			{				
				char cChar=sLine[i];

				// yeah I know its lame but its 3:30am and I had v.long debugging session :-/
				switch((int)cChar)
				{
					case 39:
					case 145:
					case 146:
					case 147:
					case 148:
						oSB.Append("&#"+((int)cChar).ToString()+";");
						continue;

					default:
						break;
				};

				if(cChar<sEntityReverseLookup.Length)
				{
		
					if(sEntityReverseLookup[(int)cChar]!=null)
					{
						oSB.Append("&");
						oSB.Append(sEntityReverseLookup[(int)cChar]);
						oSB.Append(";");
						continue;
					}
				}

				oSB.Append(cChar);
			}		

			return oSB.ToString();
		}

		/// <summary>
		/// Constructs parser object using provided HTML as source for parsing
		/// </summary>
		/// <param name="p_oHTML"></param>
		public HTMLparser(string p_oHTML)
		{
			Init(p_oHTML);
		}

		/// <summary>
		/// Sets text treatment mode 
		/// </summary>
		/// <param name="p_bTextMode">If TRUE, then text will be parsed, if FALSE then it will be ignored (parsing of tags will be faster however)</param>
		public void SetTextMode(bool p_bTextMode)
		{
			bTextMode=p_bTextMode;
		}


		/// <summary>
		/// Initialises parses with HTML to be parsed from provided string
		/// </summary>
		/// <param name="p_oHTML">String with HTML in it</param>
		public void Init(string p_oHTML)
		{
			Init(Encoding.Default.GetBytes(p_oHTML));
		}

		/// <summary>
		/// Initialises parses with HTML to be parsed from provided data buffer: this is best in terms of
		/// correctness of parsing of various encodings that can be used in HTML
		/// </summary>
		/// <param name="p_bHTML">Data buffer with HTML in it</param>
		public void Init(byte[] p_bHTML)
		{
			// set default encoding
			oEnc=Encoding.Default;

			CleanUp();
			bHTML=p_bHTML;
			iDataLength=bHTML.Length;
		}

		/// <summary>
		/// Cleans up parser in preparation for next parsing
		/// </summary>
		public void CleanUp()
		{
			if(oEntities==null)
			{
				oEntities=InitEntities(ref iMinEntityLen,ref iMaxEntityLen,out sEntityReverseLookup);
				bMiniEntities=false;
			}

			bHTML=null;

			iCurPos=0;
			iDataLength=0;
		}

		/// <summary>
		/// Resets current parsed data to start
		/// </summary>
		public void Reset()
		{
			iCurPos=0;
		}

		/// <summary>
		/// Finishes parsing of comments tag
		/// </summary>
		/// <returns>HTMLchunk object</returns>
		private HTMLchunk ParseComments()
		{
			byte cChar=0;

			while(iCurPos<iDataLength)
			{
				cChar=bHTML[iCurPos++];

				oChunk.Append(cChar);

				if(cChar==62)
				{
					if(LookBack(2)=='-' && LookBack(3)=='-')
					{
						//oChunk.Finalise();
						return oChunk;
					}
				}

			}

			//oChunk.Finalise();
			return oChunk;
		}

		/// <summary>
		/// /script sequence indicating end of script tag
		/// </summary>
		static byte[] bClosedScriptTag={ (byte)'/', (byte)'s', (byte)'c', (byte)'r', (byte)'i', (byte)'p', (byte)'t', (byte)'>'}; 

		/// <summary>
		/// Finishes parsing of data after scripts tag - makes extra checks to avoid being broken
		/// with >'s used to denote comparison
		/// </summary>
		/// <returns>HTMLchunk object</returns>
		private HTMLchunk ParseScript()
		{
			byte cChar=0;

			oChunk.iBufPos=0;

			while(iCurPos<iDataLength)
			{
				cChar=bHTML[iCurPos++];

				if(cChar=='<')
				{
					int iLastPos=iCurPos;

					int iPos=0;

					// check here if its an HTML comment

					if(iCurPos<iDataLength)
					{
						if(bHTML[iCurPos]=='!')
						{
							if((iCurPos+3)<iDataLength)
							{
								if(bHTML[iCurPos+1]=='-' && bHTML[iCurPos+2]=='-')
								{
									// FIXIT: perhaps it is more correct here to return straight away?
									ParseComments();
									continue;
									//Console.WriteLine("");
								}
							}
							
						}
					}

					while(iCurPos<iDataLength)
					{
						cChar=bHTML[iCurPos++];

						// skip whitespace
						if(cChar<=32 && bWhiteSpace[cChar]==1)
							continue;

						if(cChar>=65 && cChar<=90)
							cChar+=32;

						// if next char it out of expected sequence then we will ignore it
						// and restart scanning from previous position
						if(cChar!=bClosedScriptTag[iPos])
						{
							iCurPos=iLastPos;
							break;
						}

						// if we got whole length then we found end of script tag and can return back
						if(++iPos==bClosedScriptTag.Length)
						{
							return oChunk;
						}

					}
					
				}

				oChunk.Append(cChar);
			}

			return oChunk;
		}

		/// <summary>
		/// Internal: parses tag that started from current position
		/// </summary>
		/// <param name="bKeepWhiteSpace">If true then whitespace will be kept, if false then it won't be (faster option)</param>
		/// <returns>HTMLchunk with tag information</returns>
#if UNSAFE_CODE || true
		private unsafe HTMLchunk ParseTag(bool bKeepWhiteSpace)
#else
		private HTMLchunk ParseTag(bool bKeepWhiteSpace)
#endif
		{
			/*
			 *  WARNING: this code was optimised for performance rather than for readability, 
			 *  so be extremely careful at changing it -- your changes could easily result in wrongly parsed HTML
			 * 
			 *  This routine takes about 60% of CPU time, in theory its the best place to gain extra speed,
			 *  but I've spent plenty of time doing it, so it won't be easy... and if it is easy then please post
			 *  your changes for everyone to enjoy!
			 * 
			 * 
			 * */

			sText.Clear();

			//oChunk.Clear();
			bool bWhiteSpaceHere=false;
			
			// for tracking quotes purposes
			bool bQuotes=false;
			byte cQuotes=0x20;
			//bool bParamValue=false;
			byte cChar=0;
			byte cPeek=0;

			// if true it means we have parsed complete tag
			bool bGotTag=false;

#if UNSAFE_CODE
			string sParam="";
#endif
			int iEqualIdx=0;

			//bool bQuotesAllowed=false;

			//StringBuilder sText=new StringBuilder(128);
			//,HTMLchunk.TEXT_CAPACITY
			
			// we reach this function immediately after tag's byte (<) was
			// detected, so we need to save it in order to keep correct HTML copy
			oChunk.Append((byte)'<'); // (byte)'<'
			
			/*
			oChunk.bBuffer[0]=60;
			oChunk.iBufPos=1;
			oChunk.iHTMLen=1;
			*/

			//while(!Eof())
			//int iTagLength=0;

			// initialise peeked char
			if(iCurPos<iDataLength)
				cPeek=bHTML[iCurPos];
			else
				cPeek=0;

			//while(iCurPos<iDataLength)
			while(cPeek!=0)
			{
				// check if its whitespace -- this is statistically low probability, thus we should
				// save on loop setup and further IFs
				// NOTE: cPeek at this stage refers to CURRENT char as in the previous run it was NEXT char
				//if(cPeek==' ' || cPeek=='\t' || cPeek==13 || cPeek==10)
				if(cPeek<=32 && bWhiteSpace[cPeek]==1)
				{
					iCurPos++;
					bWhiteSpaceHere=true;

					if(!bQuotes && !bKeepWhiteSpace)
					{
						// speculative loop unroll -- we have a good chance of seeing non-space char next
						// so instead of setting up loop we will just read it directly, this should save
						// on having to prepare while() loop
						if(iCurPos<iDataLength)
							cChar=bHTML[iCurPos++];
						else
							cChar=0;

						//if(cChar==' ' || cChar=='\t' || cChar==13 || cChar==10)
						if(cChar<=32 && bWhiteSpace[cChar]==1)
						{

							while(iCurPos<iDataLength)
							{
								cChar=bHTML[iCurPos++];

								if(cChar!=' ' && cChar!='\t' && cChar!=13 && cChar!=10)
								{
									//cPeek=bHTML[iCurPos];
									break;
								}
							}

							if(iCurPos>=iDataLength)
								cChar=0;
						}
					
						//cChar=NextChar();

#if FASTEST_PARSING
						if(!bGotTag)
							oChunk.Append((byte)' ');
#else
						if(bKeepRawHTML || !bGotTag)
							oChunk.Append((byte)' ');
#endif
					}
					else
					{
						//iCurPos++;

							// we don't want that nasty unnecessary 0x0D or 13 byte :-/
						if(cPeek!=13 && (bKeepWhiteSpace || bQuotes))
						{
#if FASTEST_PARSING
							if(!bGotTag)
								oChunk.Append((byte)' ');
#else
						if(bKeepRawHTML || !bGotTag)
							oChunk.Append((byte)cPeek);
#endif

							//sText.Append(cChar);
							// NOTE: this is manual inlining from actual object
							// NOTE: speculative execution that requires large enough buffer to avoid overflowing it
							/*
								if(sText.iBufPos>=DynaString.TEXT_CAPACITY)
								{
									sText.Append(cChar);
								}
							*/
						
							//sText.Length++;
							sText.bBuffer[sText.iBufPos++]=cPeek;
						}

						if(iCurPos<iDataLength)
							cPeek=bHTML[iCurPos];
						else
							cPeek=0;
						continue;
					}

				}
				else
				{
					// reuse Peeked char from previous run
					//cChar=cPeek; iCurPos++;
					if(iCurPos<iDataLength)
						cChar=bHTML[iCurPos++];
					else
						cChar=0;

					bWhiteSpaceHere=false;
				}

				//Utils.WriteLine("Got: '{0}'",(char)cChar);

				// NOTE: we assume here that entities can't be present in TAGS that have to be ASCII
#if !FASTEST_PARSING

				if(cChar==38)
				{
					cChar=(byte)CheckForEntity();

					// restore current symbol
					if(cChar==0)
						cChar=38; //(byte)'&';
					else
					{
						// we have to skip now to next byte, since 
						// some converted chars might well be control chars like >
						oChunk.bEntities=true;

						if(cChar=='<')
							oChunk.bLtEntity=true;
						
						// unless is space we will ignore it
						// note that this won't work if &nbsp; is defined as it should
						// byte int value of 160, rather than 32.
						//if(cChar!=' ')
						oChunk.Append(cChar);

						//continue;
					}
				}
#endif

				//cPeek=bHTML[iCurPos];
				 //cPeek=Peek();

#if !FASTEST_PARSING
				if(iCurPos<iDataLength)
					cPeek=bHTML[iCurPos];
				else
					cPeek=0;
#endif					
				// check if we've got tag now: either whitespace before current symbol or the next one is end of string
				if(!bGotTag)
				{
#if FASTEST_PARSING
					// WARNING: assumption here that Init code will have added ZERO at the end of HTML stream
					// which would remove necessity to check range explicitly
					if(iCurPos<iDataLength)
						cPeek=bHTML[iCurPos];
					else
						cPeek=0;
#endif

					if((sText.iBufPos>=3 && (sText.bBuffer[0]=='!' && sText.bBuffer[1]=='-' && sText.bBuffer[2]=='-')))
					{
						// we detected start of comments here, instead of parsing the rest here we will
						// call special function tuned to do the job much more effectively
						oChunk.sTag="!--";
						oChunk.oType=HTMLchunkType.Comment;
						oChunk.bComments=true;
						oChunk.Append(cChar);
						return ParseComments();
					}



					// tags can only be ASCII, therefore we can lowercase them in place and avoid hit of 
					// having to do so when we finally found end of tag
					if(cChar>=65 && cChar<=90)
						cChar+=32;

					oChunk.Append(cChar);

					if(bWhiteSpaceHere || (!bWhiteSpaceHere && cPeek=='>')  || cPeek==0)
						
						//|| sText.sText=="!--") || sText.ToString()=="!--"))	//(sText.bBuffer[0]=='!' || sText.sText=="!--") &&
					{
						if(cPeek=='>' && cChar!='/')
							sText.Append(cChar);

						oChunk.sTag=sText.ToString();

						bGotTag=true;

						//sText.Remove(0,sText.Length);
						sText.Clear();
					}
				}
				else
				{

#if FASTEST_PARSING
					if(cChar==38)
					{
						char cEntityChar=CheckForEntity();

						// restore current symbol
						if(cEntityChar==0)
							cChar=38; //(byte)'&';
						else
						{
							// we have to skip now to next byte, since 
							// some converted chars might well be control chars like >
							oChunk.bEntities=true;

							if(cChar=='<')
								oChunk.bLtEntity=true;
						
							// unless is space we will ignore it
							// note that this won't work if &nbsp; is defined as it should
							// byte int value of 160, rather than 32.
							//if(cChar!=' ')
							oChunk.Append(cEntityChar);
					
							if(iCurPos<iDataLength)
								cPeek=bHTML[iCurPos];
							else
								cPeek=0;

							continue;

							//continue;
						}
					}
#endif
#if !FASTEST_PARSING
					if(bKeepRawHTML)
						oChunk.Append(cChar);
#endif
#if FASTEST_PARSING
					// WARNING: assumption here that Init code will have added ZERO at the end of HTML stream
					// which would remove necessity to check range explicitly
					if(iCurPos<iDataLength)
						cPeek=bHTML[iCurPos];
					else
						cPeek=0;
#endif

					// ought to be parameter
					if(sText.iBufPos!=0)
					{
						if((bWhiteSpaceHere && !bQuotes) || (cPeek==0 || (cPeek=='>' && (bQuotes || (cChar!='\"' && cChar!='\'')))))
						{
							if(cPeek=='>' && cChar!='/' && cQuotes!=cChar)
								sText.Append(cChar);
				

#if UNSAFE_CODE || true
							iEqualIdx=-1;

							// scan string to identify position of = that separates param name and param value
#if true
							fixed(byte* pBuffer=sText.bBuffer)
							{
								int iPos=0;
								for(byte *p=pBuffer; iPos<sText.iBufPos; iPos++, p++)
								{

									//if(iEqualIdx==-1)
								
									// anything before = will be lower cased in place
									// this lower casing code will work for ASCII only -- should not be
									// used for anything apart from parameters and tags -- not param values
									if(*p>=65 && *p<=90)
										*p=(byte)(*p+32);
									else
									{
										if(*p=='=')
										{
											if(iPos>0)
												oChunk.AddParam(new String((sbyte*)pBuffer,0,iPos),new String((sbyte*)p+1,0,sText.iBufPos-iPos-1));
											iEqualIdx=iPos;
											break;
										}
									}
											
								

								}
					
							}
#else							
							{
								//int iEqual=-1;
								int iPos=0;

								for(;iPos<sText.iBufPos; iPos++)
								{
									byte bChar=sText.bBuffer[iPos];

									if(iEqualIdx==-1)
									{
										// anything before = will be lower cased in place
										// this lower casing code will work for ASCII only -- should not be
										// used for anything apart from parameters and tags -- not param values
										if(bChar>=65 && bChar<=90)
										{
											sText.bBuffer[iPos]+=32;
										}
										else
										{
											if(bChar=='=')
											{
												if(iPos>0)
													oChunk.AddParam(Encoding.Default.GetString(sText.bBuffer,0,iPos),Encoding.Default.GetString(sText.bBuffer,iPos+1,sText.iBufPos-iPos-1));
												iEqualIdx=iPos;
												break;
											}
										}
											
									}
								}


							}
#endif
							if(iEqualIdx<=0)
							{
								oChunk.AddParam(sText.ToString(),"");
							}

#else
							// some params consist of key=value combination
							sParam=sText.ToString();

							iEqualIdx=sParam.IndexOf('=');

							//bQuotesAllowed=false;

							if(iEqualIdx<=0)
							{
								/*
#if UNSAFE_CODE
								Utils.FastToLower(sParam);
								oChunk.AddParam(sParam,"");
#else
								oChunk.AddParam(sParam.ToLower(),"");
#endif
*/
								oChunk.AddParam(sParam.ToLower(),"");
							}
							else
							{
								oChunk.AddParam(sParam.Substring(0,iEqualIdx).ToLower(),sParam.Substring(iEqualIdx+1,sParam.Length-iEqualIdx-1));
								//bQuotesAllowed=true;
							}
#endif						
							
							//sText.Remove(0,sText.Length);
							sText.Clear();
#if UNSAFE_CODE
							sParam=null;
#endif
						}

					}
				}

				switch((byte)cChar)
				{
					case 0:
						goto GetOut;
							
						//case (byte)'>':
					case 62:

						//bQuotesAllowed=false;
						// if we are in comments mode then we will be waiting for -->
						if(!bQuotes)
						{
							if(oChunk.bClosure)
								oChunk.oType=HTMLchunkType.CloseTag;
							else
								oChunk.oType=HTMLchunkType.OpenTag;

							return oChunk;
						}

						goto AddSymbol;

						/*
						else
						{
							// some sanity check here to avoid scanning lots of html due to misplaced quote
							goto AddSymbol;
						}

						break;
						 * */

						//case (byte)'"':
					case 34:
						//case (byte)'\'':
					case 39:

						if(bQuotes)
						{
							if(cQuotes==cChar)// && bQuotesAllowed)
							{
								bQuotes=false;
								//bQuotesAllowed=false;
							}
							else
								goto AddSymbol;
						}
						else
						{
							// some HTML is not well validated and a misplaced quote can cause real big pain - far
							// more than a small misplaced char should cause
							if(bGotTag)
							{
								// there should be at least two symbols for param = and at least one char of param itself
								if(sText.iBufPos>=2)
								{
									if(sText.bBuffer[sText.iBufPos-1]=='=')
									{
										bQuotes=true;
										cQuotes=(byte)cChar;
										continue;
									}
								}
							}


							goto AddSymbol;
						}

						break;

						//case (byte)'/':
					case 47:

						if(!bQuotes && (!bGotTag || (bGotTag && oChunk.iParams==0 && sText.iBufPos==0)))
							oChunk.bClosure=true;
						else
							goto AddSymbol;
	
						break;
						
					default:
					
						AddSymbol:
							
							//if(bWhiteSpace && bQuotes)
							//	sText.Append(0x20);

							//sText.Append(cChar);
							// NOTE: this is manual inlining from actual object

							// NOTE: we go here for speculative insertion that expectes that we won't run out of
							// buffer, which should be big enough to hold most of HTML data.
							/*
								if(sText.iBufPos>=DynaString.TEXT_CAPACITY)
								{
									sText.Append(cChar);
								}
								else
							*/
						
							//sText.Length++;
						sText.bBuffer[sText.iBufPos++]=cChar;

						break;
				};

			}

			GetOut:

			if(oChunk.bClosure)
				oChunk.oType=HTMLchunkType.CloseTag;
			else
				oChunk.oType=HTMLchunkType.OpenTag;

			return oChunk;
		}

		/// <summary>
		/// Returns true if we reached end of data
		/// </summary>
		/// <returns>True if we reached data end</returns>
		private bool Eof()
		{
			if(iCurPos>=iDataLength)
				return true;

			return false;
		}

		/// <summary>
		/// Looks back X bytes and returns char that was there, or 0 if its start
		/// </summary>
		/// <returns>Previous byte, or 0 if we will reached start position</returns>
		private byte LookBack(int iHowFar)
		{
			if(iCurPos>=iHowFar)
				return bHTML[iCurPos-iHowFar];

			return 0;
		}

		/// <summary>
		/// Puts back current byte
		/// </summary>
		private void PutChar()
		{
			if(iCurPos>0)
				iCurPos--;
		}

		/// <summary>
		/// Puts back specified number of chars (bytes really)
		/// </summary>
		/// <param name="iChars">Number of chars</param>
		private void PutChars(int iChars)
		{
			if((iCurPos-iChars)>=0)
				iCurPos-=iChars;
		}

		/// <summary>
		/// Returns next char in data and increments points
		/// </summary>
		/// <returns>Next char or 0 to indicate end of data</returns>
		private byte NextChar()
		{

			//if(Eof())
			if(iCurPos>=iDataLength)
				return 0;

			//iCurPos++;
	
			//return bCharConv ? (byte)oCharset.ConvertByte(bHTML[iCurPos-1]) : (byte)bHTML[iCurPos-1];

			return bHTML[iCurPos++];
		}

		/*
		private bool byte.IsWhiteSpace(byte cChar)
		{
			switch(cChar)
			{
				case '\n':
				case ' ':
				case '\t':
				case '\r':

					return true;

				default:

					return false;
			}
		}
		*/

		/// <summary>
		/// Skips whitespace
		/// </summary>
		/// <returns>True if any whitespace were skipped, false otherwise</returns>
		private bool SkipWhiteSpace()
		{
			byte cChar;
			bool bWhiteSpace=false;
			
			/*
			while((cChar=NextChar())!=0)
			{

				if(!char.IsWhiteSpace((char)cChar) && cChar!=13 && cChar!=10)
				{
					//PutChar();
					iCurPos--;
					return bWhiteSpace;
				}
				else
					bWhiteSpace=true;
			}
			*/
			
			while(iCurPos<iDataLength)
			{
				cChar=bHTML[iCurPos++];

				if(cChar!=' ' && cChar!='\t' && cChar!=13 && cChar!=10)
				{
					//PutChar();
					iCurPos--;
					return bWhiteSpace;
				}
				else
					bWhiteSpace=true;
			}
			
			//iCurPos++;
	
			//return bCharConv ? (byte)oCharset.ConvertByte(bHTML[iCurPos-1]) : (byte)bHTML[iCurPos-1];

			/*
			fixed(byte *bPointer=&bHTML[iCurPos])
			{
				*bPointer++=0x20;

				//return *(int*)bpPointer;
			}
			*/

			return bWhiteSpace;
		}

		
		/// <summary>
		/// Parses next chunk and returns it - whitespace is NOT kept
		/// </summary>
		/// <returns>HTMLchunk or null if end of data reached</returns>
		public HTMLchunk ParseNext()
		{
			return ParseNext(false);
		}


		/// <summary>
		/// Font sizes as described by W3C: http://www.w3.org/TR/REC-CSS2/fonts.html#propdef-font-size
		/// </summary>
		/// <exclude/>
		public enum FontSize
		{
			Small_xx	= 0,
			Small_x		= 1,
			Small		= 2,
			Medium		= 3,
			Large		= 4,
			Large_x		= 5,
			Large_xx	= 6,

			Unknown		= 7,
		}

		/// <summary>
		/// Checks if first font is bigger than the second
		/// </summary>
		/// <param name="oFont1">Font #1</param>
		/// <param name="oFont2">Font #2</param>
		/// <returns>True if Font #1 bigger than the second, false otherwise</returns>
		public static bool IsBiggerFont(FontSize oFont1,FontSize oFont2)
		{
			return (int)oFont1>(int)oFont2;
		}

		/// <summary>
		/// Checks if first font is equal or bigger than the second
		/// </summary>
		/// <param name="oFont1">Font #1</param>
		/// <param name="oFont2">Font #2</param>
		/// <returns>True if Font #1 equal or bigger than the second, false otherwise</returns>
		public static bool IsEqualOrBiggerFont(FontSize oFont1,FontSize oFont2)
		{
			return (int)oFont1>=(int)oFont2;
		}

		/// <summary>
		/// Parses font's tag size param 
		/// </summary>
		/// <param name="sSize">String value of the size param</param>
		/// <param name="iBaseFontSize">Optional base font size, use -1 if its not present</param>
		/// <param name="iCurSize"></param>
		/// <returns>Relative size of the font size or Unknown if it was not determined</returns>
		public static FontSize ParseFontSize(string sSize,FontSize oCurSize)
		{
			// TODO: read more http://www.w3.org/TR/REC-CSS2/fonts.html#propdef-font-size
			sSize=sSize.Trim();

			if(sSize.Length==0)
				return FontSize.Unknown;

			// check if its relative or absolute value
			int iSign=0;

			string sDigits="";

			int iDigits=0;

			for(int i=0; i<sSize.Length; i++)
			{
				char cChar=sSize[i];

				if(char.IsWhiteSpace(cChar))
					continue;

				switch(cChar)
				{
					case '+':

						if(iSign==0)
							iSign=1;
						else
							return FontSize.Unknown;

						break;

					case '-':

						if(iSign==0)
							iSign=-1;
						else
							return FontSize.Unknown;

						break;

					default:

						if(char.IsDigit(cChar))
						{
							iDigits++;

							if(sDigits.Length==0)
								sDigits=cChar.ToString();
							else
								sDigits+=cChar;
						}

						break;
				};
				
			}
			
			if(sDigits.Length==0 || iDigits==0)
				return FontSize.Unknown;

			int iFontSize=0;

			if(sDigits.Length>3)
				return FontSize.Unknown;

			int iSize=0;
			
			try
			{
				iSize=int.Parse(sDigits);
			}
			catch
			{
				return FontSize.Unknown;
			}

			if(iSign==0)
			{
				// absolute set
				iFontSize=iSize;
			}
			else	// relative change 
			{
				if(iSign<0)
					iFontSize=(int)oCurSize-iSize;
				else
					iFontSize=(int)oCurSize+iSize;
			}
			
			// sanity check
			if(iFontSize<0)
				iFontSize=0;
			else
				if(iFontSize>6)
					iFontSize=6;

			return (FontSize)iFontSize;
		}

		/// <summary>
		/// Parses next chunk and returns it with 
		/// </summary>
		/// <param name="bKeepWhiteSpace">If true then whitespace will be preserved (slower)</param>
		/// <returns>HTMLchunk or null if end of data reached</returns>
		public HTMLchunk ParseNext(bool bKeepWhiteSpace)
		{
			oChunk.Clear();

			bool bWhiteSpaceHere;
			byte cChar;

			while(iCurPos<iDataLength)
			{
#if !FASTEST_PARSING
				if(!bKeepWhiteSpace)
#endif
				{
					//bWhiteSpace=SkipWhiteSpace();
					
					//if(iCurPos<iDataLength)
					cChar=bHTML[iCurPos++];
					//else
					//	break;
					//cChar=cPeek;
					//iCurPos++;
					
					// statistically 75% of chars will NOT be whitespace, so we can avoid costs of setting up loop
					//if(cChar==' ' || cChar=='\t' || cChar==10 || cChar==13)
					if(cChar<=32 && bWhiteSpace[cChar]==1)
					{
						bWhiteSpaceHere=true;

						//NOTE: we speculatively expect next char NOT to be whitespace, and this
						// would help us avoid costs of setting up loop
						if(iCurPos<iDataLength)
						{
							cChar=bHTML[iCurPos++];

							//if(cChar!=' ' && cChar!='\t' && cChar!=13 && cChar!=10)
							if(bWhiteSpace[cChar]==0)
							{
								goto WhiteSpaceDone;
							}

							while(iCurPos<iDataLength)
							{
								cChar=bHTML[iCurPos++];

								//if(cChar!=' ' && cChar!='\t' && cChar!=13 && cChar!=10)
								if(bWhiteSpace[cChar]==0)
								{
									// we don't do anything because we found char that can be used down the pipeline
									// without need to look it up again
									//PutChar();
									//iCurPos--;
									goto WhiteSpaceDone;
								}
							
							}

						}

						break;
					}
					else
					{
						bWhiteSpaceHere=false;
					}

				}
#if !FASTEST_PARSING
				else
				{
					// this branch is rarely called (well, at least with my settings)
					// so we are better off setting bWhiteSpace to value here that will never be called
					// than doing the same on top of the function knowing it will be used at least once
					bWhiteSpaceHere=false;
					cChar=NextChar();

					// we are definately done
					if(cChar==0)
						break;
				}
#endif			
			WhiteSpaceDone:		

				switch((byte)cChar)
				{
						//case '<':
					case 60:


						// we may have found text bit before getting to the tag
						// in which case we need to put back tag byte and return
						// found text first, the tag will be parsed next time
						if(!(oChunk.iBufPos==0 && bWhiteSpaceHere && bIgnoreWhiteSpaceBeforeTags))
						{

							if(oChunk.iBufPos>0 || bWhiteSpaceHere)
							{
								// we will add 1 white space chars to compensate for 
								// loss of space before tag since this space often serves as a delimiter between words
								//PutChar();
								iCurPos--;
								
								oChunk.oType=HTMLchunkType.Text;

								if(bWhiteSpaceHere)
								{
									if(oChunk.iBufPos==0)
									{
										oChunk.oHTML=" ";
										return oChunk;
									}

									oChunk.Append(0x20);
								}

								// finalise chunk if text mode is not false
								if(bTextMode)
									oChunk.Finalise();

								return oChunk;
							}
						}

						//try
						oChunk=ParseTag(bKeepWhiteSpace);

					
						//                    123456
						// check for start of script
						if(oChunk.sTag.Length==6 && oChunk.sTag=="script")
						{
							if(!oChunk.bClosure)
							{
								oChunk.oType=HTMLchunkType.Script;
								oChunk=ParseScript();
							}
						}

						if(bKeepRawHTML)
							oChunk.Finalise();

						return oChunk;

					//	catch(Exception oEx)
						{
						//	Utils.WriteLine("hoe");
						//	return null;
						}
				

						/*
						 * case 179:
							Console.WriteLine("Found: {0} in {1}!",(char)cChar,oChunk.oHTML.ToString());
							break;
							*/
						
					case 13:
						break;

					case 10:
						if(bKeepWhiteSpace)
						{
							/*
							if(oChunk==null)
							{
								oChunk=new HTMLchunk(false);
								oChunk.oType=HTMLchunkType.Text;
							}
							*/

							oChunk.Append(cChar);
						}
						break;

					case 0:
						goto Exit;

					case 38: //(byte)'&':

						char cEntityChar=CheckForEntity();

						if(bTextMode)
						{

							// restore current symbol
							if(cEntityChar!=0)
							{
								oChunk.bEntities=true;

								if(cChar=='<')
									oChunk.bLtEntity=true;
								
								if(bWhiteSpaceHere)
									oChunk.Append((byte)' ');

								oChunk.Append(cEntityChar);
							}
							else
							{
						
								if(bWhiteSpaceHere)
									oChunk.Append((byte)' ');

								oChunk.Append(cChar);
							}
						}
						break;

					default:

						/*
						if(oChunk==null)
						{
							oChunk=new HTMLchunk(false);
							oChunk.oType=HTMLchunkType.Text;
						}
						*/
						if(bTextMode)
						{

							/*
							if(bReturnSplitWords)
							{
								if(bWhiteSpaceHere)
								{
									if(oChunk.iBufPos>0)
									{
										//PutChar();
										iCurPos--;

										oChunk.Finalise();
										return oChunk;
									}
								}
								else
								{
									if(char.IsPunctuation((char)cChar))
									{
										if(oChunk.iBufPos>0)
										{
											//PutChar();
											oChunk.Finalise();
											return oChunk;
										}
										else
											break;
									}
								}
							}
							else
							{
							}
							*/
							if(bWhiteSpaceHere)
								oChunk.Append((byte)' ');
							
							oChunk.Append(cChar);
						}
						
						break;
				};

			}

			Exit:

			if(oChunk.iBufPos==0)
				return null;

			// it will be null if we have not found any data

			if(bTextMode)
				oChunk.Finalise();

			oChunk.oType=HTMLchunkType.Text;

			return oChunk;
		}

		/// <summary>
		/// Parses WIDTH param and calculates width
		/// </summary>
		/// <param name="sWidth">WIDTH param from tag</param>
		/// <param name="iAvailWidth">Currently available width for relative calculations, if negative width will be returned as is</param>
		/// <param name="bRelative">Flag that will be set to true if width was relative</param>
		/// <returns>Width in pixels</returns>
		public static int CalculateWidth(string sWidth,int iAvailWidth,ref bool bRelative)
		{
			sWidth=sWidth.Trim();

			if(sWidth.Length==0)
				return iAvailWidth;

			try
			{
				// check if its relative %-t				
				bRelative=false;
				
				for(int i=0; i<sWidth.Length; i++)
				{
					if(sWidth[i]=='%')
						bRelative=true;

					if(!char.IsNumber(sWidth[i]))
					{
						sWidth=sWidth.Substring(0,i);
						break;
					}

				}

				int iValue=int.Parse(sWidth);

				if(bRelative && iAvailWidth>0)
					return iValue*iAvailWidth/100;
				else
					return iValue;

			}
			catch{}

			return iAvailWidth;
		}
		
		/// <summary>
		/// This function will be called when & is found, and it will
		/// peek forward to check if its entity, should there be a success
		/// indicated by non-zero returned, the pointer will be left at the new byte
		/// after entity
		/// </summary>
		/// <returns>Char (not byte) that corresponds to the entity or 0 if it was not entity</returns>
#if UNSAFE_CODE
		private unsafe char CheckForEntity()
#else
		private char CheckForEntity()
#endif
		{
			if(!bDecodeEntities && !bMiniEntities)
				return (char)0;

			int iChars=0;
			byte cChar;
			//string sEntity="";

			// if true it means we are getting hex or decimal value of the byte
			bool bCharCode=false;
			bool bCharCodeHex=false;

			int iEntLen=0;

			int iFrom=iCurPos;

			string sEntity;

			try
			{

				/*
				while(!Eof())
				{
					cChar=NextChar();
				*/
				while(iCurPos<iDataLength)
				{
					cChar=bHTML[iCurPos++];

					// 21/10/05: not necessary
					//if(cChar==0)
					//	break;

					if(++iChars<=2)
					{

						// the first byte for numbers should be #
						if(iChars==1)
						{
							if(cChar=='#')
							{
								iFrom++;
								bCharCode=true;
								continue;
							}
						}
						else
						{

							if(bCharCode && cChar=='x')
							{
								iFrom++;
								iEntLen--;
								bCharCodeHex=true;
							}
						}
					}

					//Console.WriteLine("Got entity end: {0}",sEntity);
					// Break on:
					// 1) ; - proper end of entity
					// 2) number 10-based entity but current byte is not a number
					//if(cChar==';' || (bCharCode && !bCharCodeHex && !char.IsNumber((char)cChar)))
					if(cChar==';' || (bCharCode && !bCharCodeHex && !(cChar>='0' && cChar<='9')))
					{
						// lets try speculative quick lookup using just first 2 chars
						// this should be successful in almost all cases thus removing need for
						// expensive creation of a string 
						if(!bCharCode && iEntLen>1)
						{
							object oChar=oEntities.GetLikelyPresentValue(bHTML[iFrom],bHTML[iFrom+1]);

							if(oChar!=null)
							{
								return (char)((int)oChar);
							}
						}

						// check if its int - this way we can avoid having to create string 
						if(bCharCode && iEntLen>0 && !bCharCodeHex)
						{
							// if mini entities mode is set then we will ignore all numerics
							if(bMiniEntities)
								break;

							return (char) ParseUInt(bHTML,iFrom,iEntLen);
						}

#if UNSAFE_CODE
						fixed(byte* pBuffer=bHTML)
						{
							sEntity=new String((sbyte*)pBuffer,iFrom,iEntLen,System.Text.Encoding.Default);
#else
					
						sEntity=System.Text.Encoding.Default.GetString(bHTML,iFrom,iEntLen);
#endif

						if(bCharCode)
						{
							// NOTE: this may fail due to wrong data format,
							// in which case we will return 0, and entity will be
							// ignored
							if(iEntLen>0)
							{
								// if mini entities mode is set then we will ignore all numerics
								if(bMiniEntities)
									break;

								int iChar;

								if(!bCharCodeHex)
								{
#if DOTNET20
									// we want to avoid exceptions if possible as they are slow
									if(!int.TryParse(sEntity,out iChar))
									{
										if(iChars>0)
											PutChars(iChars);
										return (char)(0);
									}
#else
									iChar=int.Parse(sEntity);
#endif
								}
								else
								{
#if DOTNET20
									// we want to avoid exceptions if possible as they are slow
									
									if(!int.TryParse(sEntity,System.Globalization.NumberStyles.HexNumber,null,out iChar))
									{
										if(iChars>0)
											PutChars(iChars);
										return (char)(0);
									}
#else

									iChar=int.Parse(sEntity,System.Globalization.NumberStyles.HexNumber);
#endif
								}
								
								return (char)iChar;
							}
						}
								
						if(iEntLen>=iMinEntityLen && iEntLen<=iMaxEntityLen)
						{
							object oChar=oEntities.GetLikelyPresentValue(sEntity);

							if(oChar!=null)
								return (char)((int)oChar);
						}
					}

						//break;
					

					// as soon as entity length exceed max length of entity known to us
					// we break up the loop and return nothing found

					// NOTE: removed due to entities being generally correct and this code costs 10% of CPU in this function
					
					if(iEntLen>iMaxEntityLen)
						break;
					
					iEntLen++;
				}
			}
			catch(Exception oEx)
			{
				Console.WriteLine("Entity parsing exception: "+oEx.ToString());
			}

			// if we have not found squat, then we will need to put point back
			// to where it was before this function was called
			if(iChars>0)
				PutChars(iChars);

			return (char)(0);
		}

		/// <summary>
		/// This function will decode any entities found in a string - not fast!
		/// </summary>
		/// <returns>Possibly decoded string</returns>
#if UNSAFE_CODE
		public static unsafe char DecodeEntities()
#else
		public static string DecodeEntities(string sData)
#endif
		{
			char cChar;

			StringBuilder oSB=new StringBuilder(sData.Length);

			string sEntity="";

			try
			{
				for(int i=0; i<sData.Length; i++)
				{
					cChar=sData[i];

					if(cChar!='&')
					{
						oSB.Append(cChar);
					}
					else
					{
						// if true it means we are getting hex or decimal value of the byte
						bool bCharCode=false;
						bool bCharCodeHex=false;
						int iEntLen=0;
						int iChars=0;

						int j=i+1;

						int iFrom=i+1;

						for(; j<sData.Length; j++)
						{
							cChar=sData[j];

							if(++iChars<=2)
							{

								// the first byte for numbers should be #
								if(iChars==1)
								{
									if(cChar=='#')
									{
										iFrom++;
										bCharCode=true;
										continue;
									}
								}
								else
								{

									if(bCharCode && cChar=='x')
									{
										iFrom++;
										iEntLen--;
										bCharCodeHex=true;
									}
								}
							}

							//Console.WriteLine("Got entity end: {0}",sEntity);
							// Break on:
							// 1) ; - proper end of entity
							// 2) number 10-based entity but current byte is not a number
							//if(cChar==';' || (bCharCode && !bCharCodeHex && !char.IsNumber((char)cChar)))
							if(cChar==';' || (bCharCode && !bCharCodeHex && !(cChar>='0' && cChar<='9')))
							{
								// lets try speculative quick lookup using just first 2 chars
								// this should be successful in almost all cases thus removing need for
								// expensive creation of a string 
								if(!bCharCode && iEntLen>1)
								{
									// make sure we aint at the end of string
									if(i+2<sData.Length)
									{

										object oChar=oAllEntities.GetLikelyPresentValue((byte)sData[i+1],(byte)sData[i+2]);

										if(oChar!=null)
										{
											oSB.Append((char)((int)oChar));
											break;
										}
									}

								}

								// check if its int - this way we can avoid having to create string 
								if(bCharCode && iEntLen>0 && !bCharCodeHex)
								{
									sEntity=sData.Substring(iFrom,iEntLen);
									oSB.Append((char) uint.Parse(sEntity));
									break;
								}

#if UNSAFE_CODE
						fixed(byte* pBuffer=bHTML)
						{
							sEntity=new String((sbyte*)pBuffer,iFrom,iEntLen,System.Text.Encoding.Default);
#else
					
								sEntity=sData.Substring(i+1,iEntLen);
#endif

								if(bCharCode)
								{
									// NOTE: this may fail due to wrong data format,
									// in which case we will return 0, and entity will be
									// ignored
									if(iEntLen>0)
									{
										int iChar=0;
									
										if(!bCharCodeHex) 
											iChar=int.Parse(sEntity);
										else
											iChar=int.Parse(sEntity,System.Globalization.NumberStyles.HexNumber);
								
										oSB.Append((char)iChar);
										break;
									}
								}
								
								if(iEntLen>=iAllMinEntityLen && iEntLen<=iAllMaxEntityLen)
								{
									object oChar=oAllEntities.GetLikelyPresentValue(sEntity);

									if(oChar!=null)
									{
										oSB.Append((char)((int)oChar));
										break;
									}
								}
							}

							//break;
					

							// as soon as entity length exceed max length of entity known to us
							// we break up the loop and return nothing found

							// NOTE: removed due to entities being generally correct and this code costs 10% of CPU in this function
					
							if(iEntLen>iAllMaxEntityLen)
							{
								// append char that triggered entity thingy in the first place
								oSB.Append('&');
								j=i;
								break;
							}
					
							iEntLen++;
						}

						i=j;
					}
				}
			}
			catch(Exception oEx)
			{
				Console.WriteLine("Entity parsing exception: "+oEx.ToString());

				return sData;
			}

			return oSB.ToString();
		}

		/// <summary>
		/// Multipliers for base 10 
		/// </summary>
		static uint[] iDecMultipliers={ 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000, 1000000000 };

		/// <summary>
		/// Parses an unsigned integer number from byte buffer
		/// </summary>
		/// <param name="bBuf">Buffer to parse from</param>
		/// <param name="iFrom">Start parsing from this point</param>
		/// <param name="iLength">Length of data to parse</param>
		/// <returns>Unsigned integer number</returns>
		uint ParseUInt(byte[] bBuf,int iFrom,int iLength)
		{
			uint iNum=0;

			int iOrder=0;

			for(int i=iFrom+iLength-1; i>=iFrom; i--)
			{
				byte b=bBuf[i];

				if(b<(byte)'0' || b>(byte)'9')
					break;

				//Console.WriteLine("Order={0}, Byte={1}",iOrder,b);

				iNum+=(uint) (iDecMultipliers[iOrder++]*(b-(byte)'0'));
			}

			return iNum;
		}

		/// <summary>
		/// Loads HTML from file
		/// </summary>
		/// <param name="sFileName">Full filename</param>
		public void LoadFromFile(string sFileName)
		{
			CleanUp();

			using(FileStream oFS=File.OpenRead(sFileName))
			{
				byte[] bData=new byte[oFS.Length];
	
				int iRead=oFS.Read(bData,0,bData.Length);

				if(iRead!=bData.Length)
					throw new Exception("Number of bytes read is less than expected number of bytes in a file!");

				Init(bData);
			}

			return;
		}
	
	}

}

// THE END ... is just a new beginning